#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <assert.h>

#include <linux/slab.h>
#include <linux/radix-tree.h>

#include "test.h"

static void simple_check(void)
{
	RADIX_TREE(tree, GFP_KERNEL);

	/* check "direct" data */
	item_check_absent(&tree, 0);
	item_check_absent(&tree, 1);
	item_insert(&tree, 0);
	item_check_present(&tree, 0);
	item_check_absent(&tree, 1);
	item_delete(&tree, 0);
	item_check_absent(&tree, 0);
	item_check_absent(&tree, 1);

	/* check node data */
	item_check_absent(&tree, 0);
	item_check_absent(&tree, 1);
	item_insert(&tree, 1);
	item_check_absent(&tree, 0);
	item_check_present(&tree, 1);
	item_delete(&tree, 1);
	item_check_absent(&tree, 0);
	item_check_absent(&tree, 1);
	item_kill_tree(&tree);
}

void __gang_check(unsigned long middle, long down, long up, int chunk, int hop)
{
	long idx;
	RADIX_TREE(tree, GFP_KERNEL);

	middle = 1 << 30;

	for (idx = -down; idx < up; idx++)
		item_insert(&tree, middle + idx);

	item_check_absent(&tree, middle - down - 1);
	for (idx = -down; idx < up; idx++)
		item_check_present(&tree, middle + idx);
	item_check_absent(&tree, middle + up);

	item_gang_check_present(&tree, middle - down,
			up + down, chunk, hop);
	item_full_scan(&tree, middle - down, down + up, chunk);
	item_kill_tree(&tree);
}

void gang_check(void)
{
	__gang_check(1 << 30, 128, 128, 35, 2);
	__gang_check(1 << 31, 128, 128, 32, 32);
	__gang_check(1 << 31, 128, 128, 32, 100);
	__gang_check(1 << 31, 128, 128, 17, 7);
	__gang_check(0xffff0000, 0, 65536, 17, 7);
	__gang_check(0xfffffffe, 1, 1, 17, 7);
}

void __big_gang_check(void)
{
	unsigned long start;
	int wrapped = 0;

	start = 0;
	do {
		unsigned long old_start;

//		printf("0x%08lx\n", start);
		__gang_check(start, rand() % 113 + 1, rand() % 71,
				rand() % 157, rand() % 91 + 1);
		old_start = start;
		start += rand() % 1000000;
		if (start < old_start)
			wrapped = 1;
	} while (!wrapped);
}

void big_gang_check(void)
{
	int i;

	for (i = 0; i < 1000; i++) {
		__big_gang_check();
		srand(time(0));
		printf("%d ", i);
		fflush(stdout);
	}
}

void dynamic_height_check(void)
{
	int i;
	RADIX_TREE(tree, GFP_KERNEL);
	tree_verify_min_height(&tree, 0);
	
	item_insert(&tree, 42);
	tree_verify_min_height(&tree, 42);

	item_insert(&tree, 1000000);
	tree_verify_min_height(&tree, 1000000);

	assert(item_delete(&tree, 1000000));
	tree_verify_min_height(&tree, 42);

	assert(item_delete(&tree, 42));
	tree_verify_min_height(&tree, 0);

	for (i = 0; i < 1000; i++) {
		item_insert(&tree, i);
		tree_verify_min_height(&tree, i);
	}

	i--;
	for (;;) {
		assert(item_delete(&tree, i));
		if (i == 0) {
			tree_verify_min_height(&tree, 0);
			break;
		}
		i--;
		tree_verify_min_height(&tree, i);
	}
	
	item_kill_tree(&tree);
}

int main()
{
	radix_tree_init();

	simple_check();
	printf("after simple_check: %d allocated\n", nr_allocated);
	tag_check();
	printf("after tag_check: %d allocated\n", nr_allocated);
	gang_check();
	printf("after gang_check: %d allocated\n", nr_allocated);
	dynamic_height_check();
	printf("after dynamic_height_check: %d allocated\n", nr_allocated);
	big_gang_check();
	printf("after big_gang_check: %d allocated\n", nr_allocated);
	exit(0);
}
