#include <linux/rcupdate.h>
#include <linux/list.h>

#include <stdlib.h>
#include <sys/types.h>
#include <linux/unistd.h>
#include <errno.h>
_syscall0(pid_t,gettid)
#include <pthread.h>

static int rcu_gen = 0;
static int call_rcu_times = 0;
static int fn_run_times = 0;
static pthread_mutex_t lock = PTHREAD_MUTEX_INITIALIZER;
static LIST_HEAD(pending_rcu_list);
static LIST_HEAD(rcu_list);

static int rcu_gen_max_freed = -1;

struct rcu_user {
	int rcu_gen;
	struct list_head list;
	pid_t tid;
};

static LIST_HEAD(pending_rcu_users);
static LIST_HEAD(rcu_users);

static void run_all_rcu(struct list_head *exec_list)
{
	while (!list_empty(exec_list)) {
		struct list_head *head = exec_list->next;
		struct rcu_head *rcu_head = list_entry(head, struct rcu_head, list);
		list_del(&rcu_head->list);
		assert(rcu_head->rcu_gen < rcu_gen);
		if (rcu_gen_max_freed < rcu_head->rcu_gen)
			rcu_gen_max_freed = rcu_head->rcu_gen;
		rcu_head->func(rcu_head);
		fn_run_times++;
	}
}

static void check_rcu_lists(void)
{
	LIST_HEAD(exec_rcu_list);

	if (list_empty(&pending_rcu_users)) {
		rcu_gen++;

		list_splice_init(&pending_rcu_list, &exec_rcu_list);

		list_splice_init(&rcu_users, &pending_rcu_users);
		list_splice_init(&rcu_list, &pending_rcu_list);
		if (list_empty(&pending_rcu_users))
			list_splice_init(&pending_rcu_list, &exec_rcu_list);
	}
	pthread_mutex_unlock(&lock);

	run_all_rcu(&exec_rcu_list);
}

void call_rcu(struct rcu_head *x, void (*func)(struct rcu_head *head))
{
	x->func = func;
	smp_wmb();
	pthread_mutex_lock(&lock);
	x->rcu_gen = rcu_gen;
	call_rcu_times++;
	list_add_tail(&x->list, &rcu_list);
	check_rcu_lists();
}

void rcu_read_lock(void)
{
	struct rcu_user *ru;
	ru = malloc(sizeof(struct rcu_user));
	INIT_LIST_HEAD(&ru->list);
	ru->tid = gettid();

	pthread_mutex_lock(&lock);
	ru->rcu_gen = rcu_gen;
	list_add(&ru->list, &rcu_users);
	pthread_mutex_unlock(&lock);
}

void rcu_read_unlock(void)
{

	struct rcu_user *ru;
	pid_t tid = gettid();

	pthread_mutex_lock(&lock);
	list_for_each_entry(ru, &rcu_users, list) {
		if (ru->tid == tid)
			goto found;
	}

	list_for_each_entry(ru, &pending_rcu_users, list) {
		if (ru->tid == tid)
			goto found;
	}
	BUG_ON(1);
found:
	list_del(&ru->list);
	assert(ru->rcu_gen > rcu_gen_max_freed);
	free(ru);
	check_rcu_lists();
}

void rcu_init(void)
{
}

void rcu_exit(void)
{
	printk("call_rcu called %d times, %d rcu heads freed\n",
			call_rcu_times, fn_run_times);
}
